"""
Generic processors relating to remote operations.

These can be as simple as a wrangler which converts between ID types for a remote object,
to more complex operations/algorithms such as search+match algorithms.
"""
